# Copyright 2011, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""WebSocket utilities."""

from __future__ import absolute_import
import array
import errno
import logging
import os
import re
import six
from six.moves import map
from six.moves import range
import socket
import struct
import zlib

try:
    from mod_pywebsocket import fast_masking
except ImportError:
    pass


def prepend_message_to_exception(message, exc):
    """Prepend message to the exception."""
    exc.args = (message + str(exc), )
    return


def __translate_interp(interp, cygwin_path):
    """Translate interp program path for Win32 python to run cygwin program
    (e.g. perl).  Note that it doesn't support path that contains space,
    which is typically true for Unix, where #!-script is written.
    For Win32 python, cygwin_path is a directory of cygwin binaries.

    Args:
      interp: interp command line
      cygwin_path: directory name of cygwin binary, or None
    Returns:
      translated interp command line.
    """
    if not cygwin_path:
        return interp
    m = re.match('^[^ ]*/([^ ]+)( .*)?', interp)
    if m:
        cmd = os.path.join(cygwin_path, m.group(1))
        return cmd + m.group(2)
    return interp


def get_script_interp(script_path, cygwin_path=None):
    r"""Get #!-interpreter command line from the script.

    It also fixes command path.  When Cygwin Python is used, e.g. in WebKit,
    it could run "/usr/bin/perl -wT hello.pl".
    When Win32 Python is used, e.g. in Chromium, it couldn't.  So, fix
    "/usr/bin/perl" to "<cygwin_path>\perl.exe".

    Args:
      script_path: pathname of the script
      cygwin_path: directory name of cygwin binary, or None
    Returns:
      #!-interpreter command line, or None if it is not #!-script.
    """
    fp = open(script_path)
    line = fp.readline()
    fp.close()
    m = re.match('^#!(.*)', line)
    if m:
        return __translate_interp(m.group(1), cygwin_path)
    return None


def hexify(s):
    return ' '.join(['%02x' % x for x in six.iterbytes(s)])


def get_class_logger(o):
    """Return the logging class information."""
    return logging.getLogger('%s.%s' %
                             (o.__class__.__module__, o.__class__.__name__))


def pack_byte(b):
    """Pack an integer to network-ordered byte"""
    return struct.pack('!B', b)


class NoopMasker(object):
    """A NoOp masking object.

    This has the same interface as RepeatedXorMasker but just returns
    the string passed in without making any change.
    """
    def __init__(self):
        """NoOp."""
        pass

    def mask(self, s):
        """NoOp."""
        return s


class RepeatedXorMasker(object):
    """A masking object that applies XOR on the string.

    Applies XOR on the byte string given to mask method with the masking bytes
    given to the constructor repeatedly. This object remembers the position
    in the masking bytes the last mask method call ended and resumes from
    that point on the next mask method call.
    """
    def __init__(self, masking_key):
        self._masking_key = masking_key
        self._masking_key_index = 0

    def _mask_using_swig(self, s):
        """Perform the mask via SWIG."""
        masked_data = fast_masking.mask(s, self._masking_key,
                                        self._masking_key_index)
        self._masking_key_index = ((self._masking_key_index + len(s)) %
                                   len(self._masking_key))
        return masked_data

    def _mask_using_array(self, s):
        """Perform the mask via python."""
        if isinstance(s, six.text_type):
            raise Exception(
                'Masking Operation should not process unicode strings')

        result = bytearray(s)

        # Use temporary local variables to eliminate the cost to access
        # attributes
        masking_key = [c for c in six.iterbytes(self._masking_key)]
        masking_key_size = len(masking_key)
        masking_key_index = self._masking_key_index

        for i in range(len(result)):
            result[i] ^= masking_key[masking_key_index]
            masking_key_index = (masking_key_index + 1) % masking_key_size

        self._masking_key_index = masking_key_index

        return bytes(result)

    if 'fast_masking' in globals():
        mask = _mask_using_swig
    else:
        mask = _mask_using_array


# By making wbits option negative, we can suppress CMF/FLG (2 octet) and
# ADLER32 (4 octet) fields of zlib so that we can use zlib module just as
# deflate library. DICTID won't be added as far as we don't set dictionary.
# LZ77 window of 32K will be used for both compression and decompression.
# For decompression, we can just use 32K to cover any windows size. For
# compression, we use 32K so receivers must use 32K.
#
# Compression level is Z_DEFAULT_COMPRESSION. We don't have to match level
# to decode.
#
# See zconf.h, deflate.cc, inflate.cc of zlib library, and zlibmodule.c of
# Python. See also RFC1950 (ZLIB 3.3).


class _Deflater(object):
    def __init__(self, window_bits):
        self._logger = get_class_logger(self)

        # Using the smallest window bits of 9 for generating input frames.
        # On WebSocket spec, the smallest window bit is 8. However, zlib does
        # not accept window_bit = 8.
        #
        # Because of a zlib deflate quirk, back-references will not use the
        # entire range of 1 << window_bits, but will instead use a restricted
        # range of (1 << window_bits) - 262. With an increased window_bits = 9,
        # back-references will be within a range of 250. These can still be
        # decompressed with window_bits = 8 and the 256-byte window used there.
        #
        # Similar disscussions can be found in https://crbug.com/691074
        window_bits = max(window_bits, 9)

        self._compress = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION,
                                          zlib.DEFLATED, -window_bits)

    def compress(self, bytes):
        compressed_bytes = self._compress.compress(bytes)
        self._logger.debug('Compress input %r', bytes)
        self._logger.debug('Compress result %r', compressed_bytes)
        return compressed_bytes

    def compress_and_flush(self, bytes):
        compressed_bytes = self._compress.compress(bytes)
        compressed_bytes += self._compress.flush(zlib.Z_SYNC_FLUSH)
        self._logger.debug('Compress input %r', bytes)
        self._logger.debug('Compress result %r', compressed_bytes)
        return compressed_bytes

    def compress_and_finish(self, bytes):
        compressed_bytes = self._compress.compress(bytes)
        compressed_bytes += self._compress.flush(zlib.Z_FINISH)
        self._logger.debug('Compress input %r', bytes)
        self._logger.debug('Compress result %r', compressed_bytes)
        return compressed_bytes


class _Inflater(object):
    def __init__(self, window_bits):
        self._logger = get_class_logger(self)
        self._window_bits = window_bits

        self._unconsumed = b''

        self.reset()

    def decompress(self, size):
        if not (size == -1 or size > 0):
            raise Exception('size must be -1 or positive')

        data = b''

        while True:
            data += self._decompress.decompress(self._unconsumed,
                                                max(0, size - len(data)))
            self._unconsumed = self._decompress.unconsumed_tail
            if self._decompress.unused_data:
                # Encountered a last block (i.e. a block with BFINAL = 1) and
                # found a new stream (unused_data). We cannot use the same
                # zlib.Decompress object for the new stream. Create a new
                # Decompress object to decompress the new one.
                #
                # It's fine to ignore unconsumed_tail if unused_data is not
                # empty.
                self._unconsumed = self._decompress.unused_data
                self.reset()
                if size >= 0 and len(data) == size:
                    # data is filled. Don't call decompress again.
                    break
                else:
                    # Re-invoke Decompress.decompress to try to decompress all
                    # available bytes before invoking read which blocks until
                    # any new byte is available.
                    continue
            else:
                # Here, since unused_data is empty, even if unconsumed_tail is
                # not empty, bytes of requested length are already in data. We
                # don't have to "continue" here.
                break

        if data:
            self._logger.debug('Decompressed %r', data)
        return data

    def append(self, data):
        self._logger.debug('Appended %r', data)
        self._unconsumed += data

    def reset(self):
        self._logger.debug('Reset')
        self._decompress = zlib.decompressobj(-self._window_bits)


# Compresses/decompresses given octets using the method introduced in RFC1979.


class _RFC1979Deflater(object):
    """A compressor class that applies DEFLATE to given byte sequence and
    flushes using the algorithm described in the RFC1979 section 2.1.
    """
    def __init__(self, window_bits, no_context_takeover):
        self._deflater = None
        if window_bits is None:
            window_bits = zlib.MAX_WBITS
        self._window_bits = window_bits
        self._no_context_takeover = no_context_takeover

    def filter(self, bytes, end=True, bfinal=False):
        if self._deflater is None:
            self._deflater = _Deflater(self._window_bits)

        if bfinal:
            result = self._deflater.compress_and_finish(bytes)
            # Add a padding block with BFINAL = 0 and BTYPE = 0.
            result = result + pack_byte(0)
            self._deflater = None
            return result

        result = self._deflater.compress_and_flush(bytes)
        if end:
            # Strip last 4 octets which is LEN and NLEN field of a
            # non-compressed block added for Z_SYNC_FLUSH.
            result = result[:-4]

        if self._no_context_takeover and end:
            self._deflater = None

        return result


class _RFC1979Inflater(object):
    """A decompressor class a la RFC1979.

    A decompressor class for byte sequence compressed and flushed following
    the algorithm described in the RFC1979 section 2.1.
    """
    def __init__(self, window_bits=zlib.MAX_WBITS):
        self._inflater = _Inflater(window_bits)

    def filter(self, bytes):
        # Restore stripped LEN and NLEN field of a non-compressed block added
        # for Z_SYNC_FLUSH.
        self._inflater.append(bytes + b'\x00\x00\xff\xff')
        return self._inflater.decompress(-1)


class DeflateSocket(object):
    """A wrapper class for socket object to intercept send and recv to perform
    deflate compression and decompression transparently.
    """

    # Size of the buffer passed to recv to receive compressed data.
    _RECV_SIZE = 4096

    def __init__(self, socket):
        self._socket = socket

        self._logger = get_class_logger(self)

        self._deflater = _Deflater(zlib.MAX_WBITS)
        self._inflater = _Inflater(zlib.MAX_WBITS)

    def recv(self, size):
        """Receives data from the socket specified on the construction up
        to the specified size. Once any data is available, returns it even
        if it's smaller than the specified size.
        """

        # TODO(tyoshino): Allow call with size=0. It should block until any
        # decompressed data is available.
        if size <= 0:
            raise Exception('Non-positive size passed')
        while True:
            data = self._inflater.decompress(size)
            if len(data) != 0:
                return data

            read_data = self._socket.recv(DeflateSocket._RECV_SIZE)
            if not read_data:
                return b''
            self._inflater.append(read_data)

    def sendall(self, bytes):
        self.send(bytes)

    def send(self, bytes):
        self._socket.sendall(self._deflater.compress_and_flush(bytes))
        return len(bytes)


# vi:sts=4 sw=4 et



#!/usr/bin/python
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/. */

"""
mac_register_font.py

Mac-specific utility command to register a font file with the OS.
"""

import argparse
import sys

import Cocoa
import CoreText


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="print verbose registration failures",
        default=False,
    )
    parser.add_argument(
        "file", nargs="*", help="font file to register or unregister", default=[]
    )
    parser.add_argument(
        "-u",
        "--unregister",
        action="store_true",
        help="unregister the provided fonts",
        default=False,
    )
    parser.add_argument(
        "-p",
        "--persist-user",
        action="store_true",
        help="permanently register the font",
        default=False,
    )

    args = parser.parse_args()

    if args.persist_user:
        scope = CoreText.kCTFontManagerScopeUser
        scopeDesc = "user"
    else:
        scope = CoreText.kCTFontManagerScopeSession
        scopeDesc = "session"

    failureCount = 0
    for fontPath in args.file:
        fontURL = Cocoa.NSURL.fileURLWithPath_(fontPath)
        (result, error) = register_or_unregister_font(fontURL, args.unregister, scope)
        if result:
            print(
                "%sregistered font %s with %s scope"
                % (("un" if args.unregister else ""), fontPath, scopeDesc)
            )
        else:
            print(
                "Failed to %sregister font %s with %s scope"
                % (("un" if args.unregister else ""), fontPath, scopeDesc)
            )
            if args.verbose:
                print(error)
            failureCount += 1

    sys.exit(failureCount)


def register_or_unregister_font(fontURL, unregister, scope):
    return (
        CoreText.CTFontManagerUnregisterFontsForURL(fontURL, scope, None)
        if unregister
        else CoreText.CTFontManagerRegisterFontsForURL(fontURL, scope, None)
    )


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""A pywebsocket3 handler which runs arbitrary Python code and returns the
result.
This is used to test OS specific accessibility APIs which can't be tested in JS.
It is intended to be called from JS browser tests.
"""

import math
import os
import sys
import traceback

from mod_pywebsocket import msgutil
from mozfile import json


def web_socket_do_extra_handshake(request):
    pass


def web_socket_transfer_data(request):
    def send(*args):
        """Send a response to the client as a JSON array."""
        msgutil.send_message(request, json.dumps(args))

    cleanNamespace = {}
    testDir = None
    if sys.platform == "win32":
        testDir = "windows"
    elif sys.platform == "linux":
        testDir = "atk"
    if testDir:
        sys.path.append(
            os.path.join(
                os.getcwd(), "browser", "accessible", "tests", "browser", testDir
            )
        )
        try:
            import a11y_setup

            a11y_setup.setup()
            cleanNamespace = a11y_setup.__dict__
            setupExc = None
        except Exception:
            setupExc = traceback.format_exc()
        sys.path.pop()

    def info(message):
        """Log an info message."""
        send("info", str(message))

    cleanNamespace["info"] = info
    namespace = cleanNamespace.copy()

    # Keep handling messages until the WebSocket is closed.
    while True:
        code = msgutil.receive_message(request)
        if not code:
            return
        if code == "__reset__":
            namespace = cleanNamespace.copy()
            send("return", None)
            continue

        if setupExc:
            # a11y_setup failed. Report an exception immediately.
            send("exception", setupExc)
            continue

        # Wrap the code in a function called run(). This allows the code to
        # return a result by simply using the return statement.
        if "\n" not in code and not code.lstrip().startswith("return "):
            # Single line without return. Assume this is an expression. We use
            # a lambda to return the result.
            code = f"run = lambda: {code}"
        else:
            lines = ["def run():"]
            # Indent each line inside the function.
            lines.extend(f"  {line}" for line in code.splitlines())
            code = "\n".join(lines)
        try:
            # Execute this Python code, which will define the run() function.
            exec(code, namespace)
            # Run the function we just defined.
            ret = namespace["run"]()
            if isinstance(ret, float) and math.isnan(ret):
                # NaN can't be serialized by JSON.
                ret = None
            send("return", ret)
        except Exception:
            send("exception", traceback.format_exc())



#! /usr/bin/env python2

"""generates tests from OpenGL ES 2.0 .run/.test files."""

import os
import os.path
import sys
import re
import json
import shutil
from optparse import OptionParser
from xml.dom.minidom import parse

if sys.version < '2.6':
   print 'Wrong Python Version !!!: Need >= 2.6'
   sys.exit(1)

# each shader test generates up to 3 512x512 images.
# a 512x512 image takes 1meg of memory so set this
# number apporpriate for the platform with
# the smallest memory issue. At 8 that means
# at least 24 meg is needed to run the test.
MAX_TESTS_PER_SET = 8

VERBOSE = False

FILTERS = [
  re.compile("GL/"),
]

LICENSE = """
Copyright (c) 2019 The Khronos Group Inc.
Use of this source code is governed by an MIT-style license that can be
found in the LICENSE.txt file.
""".strip()

COMMENT_RE = re.compile("/\*\n\*\*\s+Copyright.*?\*/",
                        re.IGNORECASE | re.DOTALL)
REMOVE_COPYRIGHT_RE = re.compile("\/\/\s+Copyright.*?\n",
                                 re.IGNORECASE | re.DOTALL)
MATRIX_RE = re.compile("Matrix(\\d)")

VALID_UNIFORM_TYPES = [
  "uniform1f",
  "uniform1fv",
  "uniform1fv",
  "uniform1i",
  "uniform1iv",
  "uniform1iv",
  "uniform2f",
  "uniform2fv",
  "uniform2fv",
  "uniform2i",
  "uniform2iv",
  "uniform2iv",
  "uniform3f",
  "uniform3fv",
  "uniform3fv",
  "uniform3i",
  "uniform3iv",
  "uniform3iv",
  "uniform4f",
  "uniform4fv",
  "uniform4fv",
  "uniform4i",
  "uniform4iv",
  "uniform4ivy",
  "uniformMatrix2fv",
  "uniformMatrix2fv",
  "uniformMatrix3fv",
  "uniformMatrix3fv",
  "uniformMatrix4fv",
  "uniformMatrix4fv",
]

SUBSTITUTIONS = [
  ("uniformmat3fv", "uniformMatrix3fv"),
  ("uniformmat4fv", "uniformMatrix4fv"),
]


def Log(msg):
  global VERBOSE
  if VERBOSE:
    print msg


def TransposeMatrix(values, dim):
  size = dim * dim
  count = len(values) / size
  for m in range(0, count):
    offset = m * size
    for i in range(0, dim):
      for j in range(i + 1, dim):
        t = values[offset + i * dim + j]
        values[offset + i * dim + j] = values[offset + j * dim + i]
        values[offset + j * dim + i] = t


def GetValidTypeName(type_name):
  global VALID_UNIFORM_TYPES
  global SUBSTITUTIONS
  for subst in SUBSTITUTIONS:
    type_name = type_name.replace(subst[0], subst[1])
  if not type_name in VALID_UNIFORM_TYPES:
    print "unknown type name: ", type_name
    raise SyntaxError
  return type_name


def WriteOpen(filename):
  dirname = os.path.dirname(filename)
  if len(dirname) > 0 and not os.path.exists(dirname):
    os.makedirs(dirname)
  return open(filename, "wb")


class TxtWriter():
  def __init__(self, filename):
    self.filename = filename
    self.lines = []

  def Write(self, line):
    self.lines.append(line)

  def Close(self):
    if len(self.lines) > 0:
      Log("Writing: %s" % self.filename)
      f = WriteOpen(self.filename)
      f.write("# this file is auto-generated. DO NOT EDIT.\n")
      f.write("".join(self.lines))
      f.close()


def ReadFileAsLines(filename):
  f = open(filename, "r")
  lines = f.readlines()
  f.close()
  return [line.strip() for line in lines]


def ReadFile(filename):
  f = open(filename, "r")
  content = f.read()
  f.close()
  return content.replace("\r\n", "\n")


def Chunkify(list, chunk_size):
  """divides an array into chunks of chunk_size"""
  return [list[i:i + chunk_size] for i in range(0, len(list), chunk_size)]


def GetText(nodelist):
  """Gets the text of from a list of nodes"""
  rc = []
  for node in nodelist:
    if node.nodeType == node.TEXT_NODE:
      rc.append(node.data)
  return ''.join(rc)


def GetElementText(node, name):
  """Gets the text of an element"""
  elements = node.getElementsByTagName(name)
  if len(elements) > 0:
    return GetText(elements[0].childNodes)
  else:
    return None


def GetBoolElement(node, name):
  text = GetElementText(node, name)
  return text.lower() == "true"


def GetModel(node):
  """Gets the model"""
  model = GetElementText(node, "model")
  if model and len(model.strip()) == 0:
    elements = node.getElementsByTagName("model")
    if len(elements) > 0:
      model = GetElementText(elements[0], "filename")
  return model


def RelativizePaths(base, paths, template):
  """converts paths to relative paths"""
  rels = []
  for p in paths:
    #print "---"
    #print "base: ", os.path.abspath(base)
    #print "path: ", os.path.abspath(p)
    relpath = os.path.relpath(os.path.abspath(p), os.path.dirname(os.path.abspath(base))).replace("\\", "/")
    #print "rel : ", relpath
    rels.append(template % relpath)
  return "\n".join(rels)


def CopyFile(filename, src, dst):
  s = os.path.abspath(os.path.join(os.path.dirname(src), filename))
  d = os.path.abspath(os.path.join(os.path.dirname(dst), filename))
  dst_dir = os.path.dirname(d)
  if not os.path.exists(dst_dir):
    os.makedirs(dst_dir)
  shutil.copyfile(s, d)


def CopyShader(filename, src, dst):
  s = os.path.abspath(os.path.join(os.path.dirname(src), filename))
  d = os.path.abspath(os.path.join(os.path.dirname(dst), filename))
  text = ReadFile(s)
  # By agreement with the Khronos OpenGL working group we are allowed
  # to open source only the .vert and .frag files from the OpenGL ES 2.0
  # conformance tests. All other files from the OpenGL ES 2.0 conformance
  # tests are not included.
  marker = "insert-copyright-here"
  new_text = COMMENT_RE.sub(marker, text)
  if new_text == text:
    print "no matching license found:", s
    raise RuntimeError
  new_text = REMOVE_COPYRIGHT_RE.sub("", new_text)
  glsl_license = '/*\n' + LICENSE + '\n*/'
  new_text = new_text.replace(marker, glsl_license)
  f = WriteOpen(d)
  f.write(new_text)
  f.close()


def IsOneOf(string, regexs):
  for regex in regexs:
    if re.match(regex, string):
      return True
  return False


def CheckForUnknownTags(valid_tags, node, depth=1):
  """do a hacky check to make sure we're not missing something."""
  for child in node.childNodes:
    if child.localName and not IsOneOf(child.localName, valid_tags[0]):
      print "unsupported tag:", child.localName
      print "depth:", depth
      raise SyntaxError
    else:
      if len(valid_tags) > 1:
        CheckForUnknownTags(valid_tags[1:], child, depth + 1)


def IsFileWeWant(filename):
  for f in FILTERS:
    if f.search(filename):
      return True
  return False


class TestReader():
  """class to read and parse tests"""

  def __init__(self, basepath):
    self.tests = []
    self.modes = {}
    self.patterns = {}
    self.basepath = basepath

  def Print(self, msg):
    if self.verbose:
      print msg

  def MakeOutPath(self, filename):
    relpath = os.path.relpath(os.path.abspath(filename), os.path.dirname(os.path.abspath(self.basepath)))
    return relpath

  def ReadTests(self, filename):
    """reads a .run file and parses."""
    Log("reading %s" % filename)
    outname = self.MakeOutPath(filename + ".txt")
    f = TxtWriter(outname)
    dirname = os.path.dirname(filename)
    lines = ReadFileAsLines(filename)
    count = 0
    tests_data = []
    for line in lines:
      if len(line) > 0 and not line.startswith("#"):
        fname = os.path.join(dirname, line)
        if line.endswith(".run"):
          if self.ReadTests(fname):
            f.Write(line + ".txt\n")
            count += 1
        elif line.endswith(".test"):
          tests_data.extend(self.ReadTest(fname))
        else:
          print "Error in %s:%d:%s" % (filename, count, line)
          raise SyntaxError()
    if len(tests_data):
      global MAX_TESTS_PER_SET
      sets = Chunkify(tests_data, MAX_TESTS_PER_SET)
      id = 1
      for set in sets:
        suffix = "_%03d_to_%03d" % (id, id + len(set) - 1)
        test_outname = self.MakeOutPath(filename + suffix + ".html")
        if os.path.basename(test_outname).startswith("input.run"):
          dname = os.path.dirname(test_outname)
          folder_name = os.path.basename(dname)
          test_outname = os.path.join(dname, folder_name + suffix + ".html")
        self.WriteTests(filename, test_outname, {"tests":set})
        f.Write(os.path.basename(test_outname) + "\n")
        id += len(set)
      count += 1
    f.Close()
    return count

  def ReadTest(self, filename):
    """reads a .test file and parses."""
    Log("reading %s" % filename)
    dom = parse(filename)
    tests = dom.getElementsByTagName("test")
    tests_data = []
    outname = self.MakeOutPath(filename + ".html")
    for test in tests:
      if not IsFileWeWant(filename):
        self.CopyShaders(test, filename, outname)
      else:
        test_data = self.ProcessTest(test, filename, outname, len(tests_data))
        if test_data:
          tests_data.append(test_data)
    return tests_data

  def ProcessTest(self, test, filename, outname, id):
    """Process a test"""
    mode = test.getAttribute("mode")
    pattern = test.getAttribute("pattern")
    self.modes[mode] = 1
    self.patterns[pattern] = 1
    Log ("%d: mode: %s pattern: %s" % (id, mode, pattern))
    method = getattr(self, 'Process_' + pattern)
    test_data = method(test, filename, outname)
    if test_data:
      test_data["pattern"] = pattern
    return test_data

  def WriteTests(self, filename, outname, tests_data):
    Log("Writing %s" % outname)
    template = """<!DOCTYPE html>
<!-- this file is auto-generated. DO NOT EDIT. -->
%(license)s
<html>
<head>
<meta charset="utf-8">
<title>WebGL GLSL conformance test: %(title)s</title>
%(css)s
%(scripts)s
</head>
<body>
<canvas id="example" width="500" height="500" style="width: 16px; height: 16px;"></canvas>
<div id="description"></div>
<div id="console"></div>
</body>
<script>
"use strict";
OpenGLESTestRunner.run(%(tests_data)s);
var successfullyParsed = true;
</script>
</html>
"""
    css = [
      "../../resources/js-test-style.css",
      "../../resources/ogles-tests.css",
    ]
    scripts = [
      "../../resources/js-test-pre.js",
      "../../resources/webgl-test-utils.js",
      "ogles-utils.js",
    ]
    css_html = RelativizePaths(outname, css, '<link rel="stylesheet" href="%s" />')
    scripts_html = RelativizePaths(outname, scripts, '<script src="%s"></script>')

    html_license = '<!--\n' + LICENSE + '\n-->'
    f = WriteOpen(outname)
    f.write(template % {
        "license": html_license,
        "css": css_html,
        "scripts": scripts_html,
        "title": os.path.basename(outname),
        "tests_data": json.dumps(tests_data, indent=2)
      })
    f.close()


  def CopyShaders(self, test, filename, outname):
    """For tests we don't actually support yet, at least copy the shaders"""
    shaders = test.getElementsByTagName("shader")
    for shader in shaders:
      for name in ["vertshader", "fragshader"]:
        s = GetElementText(shader, name)
        if s and s != "empty":
          CopyShader(s, filename, outname)

  #
  # pattern handlers.
  #

  def Process_compare(self, test, filename, outname):
    global MATRIX_RE

    valid_tags = [
      ["shader", "model", "glstate"],
      ["uniform", "vertshader", "fragshader", "filename", "depthrange"],
      ["name", "count", "transpose", "uniform*", "near", "far"],
    ]
    CheckForUnknownTags(valid_tags, test)

    # parse the test
    shaders = test.getElementsByTagName("shader")
    shaderInfos = []
    for shader in shaders:
      v = GetElementText(shader, "vertshader")
      f = GetElementText(shader, "fragshader")
      CopyShader(v, filename, outname)
      CopyShader(f, filename, outname)
      info = {
        "vertexShader": v,
        "fragmentShader": f,
      }
      shaderInfos.append(info)
      uniformElems = shader.getElementsByTagName("uniform")
      if len(uniformElems) > 0:
        uniforms = {}
        info["uniforms"] = uniforms
        for uniformElem in uniformElems:
          uniform = {"count": 1}
          for child in uniformElem.childNodes:
            if child.localName == None:
              pass
            elif child.localName == "name":
              uniforms[GetText(child.childNodes)] = uniform
            elif child.localName == "count":
              uniform["count"] = int(GetText(child.childNodes))
            elif child.localName == "transpose":
              uniform["transpose"] = (GetText(child.childNodes) == "true")
            else:
              if "type" in uniform:
                print "utype was:", uniform["type"], " found ", child.localName
                raise SyntaxError
              type_name = GetValidTypeName(child.localName)
              uniform["type"] = type_name
              valueText = GetText(child.childNodes).replace(",", " ")
              uniform["value"] = [float(t) for t in valueText.split()]
              m = MATRIX_RE.search(type_name)
              if m:
                # Why are these backward from the API?!?!?
                TransposeMatrix(uniform["value"], int(m.group(1)))
    data = {
      "name": os.path.basename(outname),
      "model": GetModel(test),
      "referenceProgram": shaderInfos[1],
      "testProgram": shaderInfos[0],
    }
    gl_states = test.getElementsByTagName("glstate")
    if len(gl_states) > 0:
      state = {}
      data["state"] = state
      for gl_state in gl_states:
        for state_name in gl_state.childNodes:
          if state_name.localName:
            values = {}
            for field in state_name.childNodes:
              if field.localName:
                values[field.localName] = GetText(field.childNodes)
            state[state_name.localName] = values
    return data

  def Process_shaderload(self, test, filename, outname):
    """no need for shaderload tests"""
    self.CopyShaders(test, filename, outname)

  def Process_extension(self, test, filename, outname):
    """no need for extension tests"""
    self.CopyShaders(test, filename, outname)

  def Process_createtests(self, test, filename, outname):
    Log("createtests Not implemented:  %s" % filename)
    self.CopyShaders(test, filename, outname)

  def Process_GL2Test(self, test, filename, outname):
    Log("GL2Test Not implemented:  %s" % filename)
    self.CopyShaders(test, filename, outname)

  def Process_uniformquery(self, test, filename, outname):
    Log("uniformquery Not implemented:  %s" % filename)
    self.CopyShaders(test, filename, outname)

  def Process_egl_image_external(self, test, filename, outname):
    """no need for egl_image_external tests"""
    self.CopyShaders(test, filename, outname)

  def Process_dismount(self, test, filename, outname):
    Log("dismount Not implemented:  %s" % filename)
    self.CopyShaders(test, filename, outname)

  def Process_build(self, test, filename, outname):
    """don't need build tests"""
    valid_tags = [
      ["shader", "compstat", "linkstat"],
      ["vertshader", "fragshader"],
    ]
    CheckForUnknownTags(valid_tags, test)

    shader = test.getElementsByTagName("shader")
    if not shader:
      return None
    vs = GetElementText(shader[0], "vertshader")
    fs = GetElementText(shader[0], "fragshader")
    if vs and vs != "empty":
      CopyShader(vs, filename, outname)
    if fs and fs != "empty":
      CopyShader(fs, filename, outname)
    data = {
      "name": os.path.basename(outname),
      "compstat": bool(GetBoolElement(test, "compstat")),
      "linkstat": bool(GetBoolElement(test, "linkstat")),
      "testProgram": {
        "vertexShader": vs,
        "fragmentShader": fs,
      },
    }
    attach = test.getElementsByTagName("attach")
    if len(attach) > 0:
      data["attachError"] = GetElementText(attach[0], "attacherror")
    return data

  def Process_coverage(self, test, filename, outname):
    Log("coverage Not implemented:  %s" % filename)
    self.CopyShaders(test, filename, outname)

  def Process_attributes(self, test, filename, outname):
    Log("attributes Not implemented:  %s" % filename)
    self.CopyShaders(test, filename, outname)

  def Process_fixed(self, test, filename, outname):
    """no need for fixed function tests"""
    self.CopyShaders(test, filename, outname)


def main(argv):
  """This is the main function."""
  global VERBOSE

  parser = OptionParser()
  parser.add_option(
      "-v", "--verbose", action="store_true",
      help="prints more output.")

  (options, args) = parser.parse_args(args=argv)

  if len(args) < 1:
    pass # fix me

  os.chdir(os.path.dirname(__file__) or '.')

  VERBOSE = options.verbose

  filename = args[0]
  test_reader = TestReader(filename)
  test_reader.ReadTests(filename)


if __name__ == '__main__':
  sys.exit(main(sys.argv[1:]))
